/* DMA.CPP - Programmed by Jonathan Hobson */
/* 180504: Maarten Hofman: Added some debugging code */
/* 100704: Maarten Hofman: Added more debugging code */
/* 250704: Maarten Hofman: Messing with page sizes */
/* 260704: Maarten Hofman: Added alignment after new page in prep tag */
/* 280704: Maarten Hofman: Played with the number of NOPs for path 3 */
/* 270904: Maarten Hofman: Added redo_from_start for static chains */

#define PAGE_SIZE 1024

#include "dma.h"

int direct=0;

CVIFStaticDMA VIFStaticDMA_SingletonInit;
CVIFDynamicDMA VIFDynamicDMA_SingletonInit;

CGIFStaticDMA GIFStaticDMA_SingletonInit;
CGIFDynamicDMA GIFDynamicDMA_SingletonInit;

void CDMAChain::PrintPacket()
{
	SPS2Manager.FlushCache();

	unsigned int * pMem = (unsigned int *)m_pPages[0].GetCached();
	
	printf("            ___________________________________________ \n");
	printf("  Address  |127     96|95      64|63      32|31       0| Virt Addr\n");
	printf(".----------+----------+----------+----------+----------|----------.\n");

	int iWritten = ((int)m_pPtr - (int)pMem + 0xF) >> 4;

	for(int i = 0; i < iWritten; i++)
	{
		static int iPhysAddr = 0;

		if((i % 256) == 0)
		{
			iPhysAddr = m_pPages[(i / 256)].GetPhysicalAddr();
		}

		printf("| %.8X | ", iPhysAddr + (i % 256) << 4);
	
		for(int j = 0; j < 4; j++)
		{
			printf("%.8X | ", pMem[3 - j]);
		}

		printf("%.8X | ", (int)pMem);

		printf("\n");
		pMem += 4;
	}

	printf("'-----------------------------------------------------------------'\n");
}

void CDMAChain::Stitch() {
bool wasdirect;

        wasdirect=false;
	if(m_VifState == VIF_DIRECT) {
          EndDirect();
          wasdirect=true;
        }

	*m_pDMATag = ((uint64)NewPage() << 32) + (2 << 28) + (m_pPtr - (int *)m_pDMATag) / 4 - 1;
	NewTag();

	if(wasdirect) StartDirect();
}

void CDMAChain::NewTag()
{
	assert(m_pPtr != m_pEndPtr);
	m_pDMATag = (uint64 *)m_pPtr;
	Add64(0);
}

void CDMAChain::PrepForDMATag() {
	assert(m_VifState == VIF_BASE);
	Align(4);
	if(m_pPtr >= m_pEndPtr) {
	  Add32(NOP);
          Align(4);
        }
}

void CDMAChain::DMAEnd()
{
	PrepForDMATag();
	*m_pDMATag = (7 << 28) + (m_pPtr - (int *)m_pDMATag) / 4 - 1;
}

void CDMAChain::DMARet()
{
	PrepForDMATag();
	*m_pDMATag = (6 << 28) + (m_pPtr - (int *)m_pDMATag) / 4 - 1;
	NewTag();
}

void CDMAChain::DMACall(uint32 iAddr)
{
	PrepForDMATag();
	*m_pDMATag = ((uint64)iAddr << 32) + (5 << 28) + (m_pPtr - (int *)m_pDMATag) / 4 - 1;
	NewTag();
}

void CDMAChain::StartDirect()
{
	assert(m_VifState == VIF_BASE);
        direct++;
	Align(4, 3);
	m_VifState = VIF_DIRECT;
	m_pVIFCode = m_pPtr;
	Add32(0);
}

void CDMAChain::EndDirect()
{
	assert(m_VifState == VIF_DIRECT);
        direct--;
	Align(4);
	int iSize = (m_pPtr - m_pVIFCode - 1) / 4;
	if(iSize != 0)
		*m_pVIFCode = (0x50 << 24) + iSize;
	m_VifState = VIF_BASE;
}

void CDMAChain::StartAD()
{
	if(m_pPtr >= m_pEndPtr)
	{
		EndDirect();
		StartDirect();
	}

	m_pADGiftag = (uint64 *)m_pPtr;
	Add64((1 << 15) | (1ULL << 60));
	Add64(0xE);
}

void CDMAChain::AddAD(uint64 iData, uint64 iAddr)
{
	Add64(iData);
	Add64(iAddr);
	(*m_pADGiftag)++;
}

void CDMAChain::StartMPG(uint32 iAddr)
{
	assert(m_VifState == VIF_BASE);
	Align(2, 1);
	m_pVIFCode = m_pPtr;
	Add32((0x4a << 24) + iAddr);
	m_iMPGCount = 0;
	m_iMPGAddr = iAddr;
}

void CDMAChain::EndMPG()
{
	*m_pVIFCode += (m_iMPGCount & 0xFF) << 16;
}

void CDMAChain::AddMPG(uint64 iInst)
{
	if(m_iMPGCount >= 256)
	{
		assert(m_iMPGCount == 256);
		EndMPG();
		StartMPG(m_iMPGAddr + m_iMPGCount);
	}
	Add64(iInst);
	m_iMPGCount++;
}

void CDMAChain::Align(int iAlign, int iOffset = 0)
{
	int p = (((int)m_pPtr >> 2) - iOffset) & (iAlign - 1);
	if(p == 0)
		return;
	for(p = iAlign - p; p > 0; p--)
		Add32(0);
}

void CDMAChain::AddVectorF(float x, float y, float z, float w)
{
	AddFloat(x);	
	AddFloat(y);
	AddFloat(z);
	AddFloat(w);
}

void CDMAChain::AddVectorI(int x, int y, int z, int w)
{
	Add32(x);
	Add32(y);
	Add32(z);
	Add32(w);
}

void CDMAChain::AddMatrix(const KazMaths::Matrix4x4 & Matrix)
{
	AddFloat(Matrix(0, 0));
	AddFloat(Matrix(0, 1));
	AddFloat(Matrix(0, 2));
	AddFloat(Matrix(0, 3));

	AddFloat(Matrix(1, 0));
	AddFloat(Matrix(1, 1));
	AddFloat(Matrix(1, 2));
	AddFloat(Matrix(1, 3));

	AddFloat(Matrix(2, 0));
	AddFloat(Matrix(2, 1));
	AddFloat(Matrix(2, 2));
	AddFloat(Matrix(2, 3));

	AddFloat(Matrix(3, 0));
	AddFloat(Matrix(3, 1));
	AddFloat(Matrix(3, 2));
	AddFloat(Matrix(3, 3));
}

void CDMAChain::Begin()
{
	m_VifState = VIF_BASE;
	m_iPhysAddr = NewPage();
	NewTag();
}

void CDMAChainNoTTE::NewTag()
{
	assert(m_pPtr != m_pEndPtr);
	m_pDMATag = (uint64 *)m_pPtr;
	Add128(0);
}

CVIFDynamicDMA::CVIFDynamicDMA()
{
	m_iCurrPage = 0;
	m_iCurrBuffer = 0;
}

CVIFDynamicDMA::~CVIFDynamicDMA()
{

}

void CVIFDynamicDMA::OpenPath3()
{
	DMACall(VIFStaticDMA.m_iPath3SyncCodes);
}

void CVIFDynamicDMA::Initialise(int iNumPages)
{
	m_iNumPages = iNumPages;

	// We are using a double buffer so allocate 2x the memory
	m_pPages = new CDMAMem[m_iNumPages * 2];

	for(int iPage = 0; iPage < (m_iNumPages * 2); iPage++)
		SPS2Manager.Allocate(m_pPages[iPage], 256);

	*EE_VIF1_ERR = 2; // Ignore DMA mismatch error

	Begin();
}

int CVIFDynamicDMA::NewPage() {
	assert(m_iCurrPage < m_iNumPages);

	m_pPtr = (int *)m_pPages[m_iCurrPage + (m_iCurrBuffer * m_iNumPages)].GetCached();
	m_pEndPtr = m_pPtr + PAGE_SIZE;
	int iPhysAddr = (int)m_pPages[m_iCurrPage + (m_iCurrBuffer * m_iNumPages)].GetPhysicalAddr();
	++m_iCurrPage;

	return iPhysAddr;
}

void CVIFDynamicDMA::PrintPacket()
{
	SPS2Manager.FlushCache();

	unsigned int * pMem = (unsigned int *)m_pPages[(m_iCurrBuffer * m_iNumPages)].GetCached();
	
	printf("            ___________________________________________            \n");
	printf("  Address  |127     96|95      64|63      32|31       0| Virt Addr \n");
	printf(".----------+----------+----------+----------+----------|----------.\n");

	int iWritten = ((int)m_pPtr - (int)pMem + 0xF) >> 4;

	for(int i = 0; i < iWritten; i++)
	{
		static int iPhysAddr = 0;

		if((i % 256) == 0)
		{
			iPhysAddr = m_pPages[(i / 256) + (m_iCurrBuffer * m_iNumPages)].GetPhysicalAddr();
		}

		printf("| %.8X | ", iPhysAddr + (i % 256) << 4);
	
		for(int j = 0; j < 4; j++)
		{
			printf("%.8X | ", pMem[3 - j]);
		}

		printf("%.8X | ", (int)pMem);

		printf("\n");
		pMem += 4;
	}

	printf("'-----------------------------------------------------------------'\n");
}

void CVIFDynamicDMA::Fire(bool bFire = true, bool bPrint = false)
{
	DMAEnd();

	if(bPrint)
		PrintPacket();

	asm("sync.l");
	SPS2Manager.FlushCache();

        if(bPrint) fprintf(stderr,"After FlushCache\n");
	
	while(*EE_D1_CHCR & 256);

        if(bPrint) fprintf(stderr,"After while()\n");

	*EE_D1_QWC = 0;
	*EE_D1_TADR = m_iPhysAddr & 0x8ffffff0;

	if(bFire)
		*EE_D1_CHCR = 0x145;

	asm("sync.l");

        if(bPrint) fprintf(stderr,"After second sync.l and bFire\n");

	m_iCurrPage = 0;
	m_iCurrBuffer ^= 1;

	Begin();
}

CVIFStaticDMA::CVIFStaticDMA()
{
	m_iCurrPage = 0;
}

CVIFStaticDMA::~CVIFStaticDMA()
{

}

void CVIFStaticDMA::redo_from_start() {
int vifcode;

  m_iCurrPage=0;
  *EE_VIF1_ERR=2;
  Begin();
  m_iPath3SyncCodes = VIFStaticDMA.GetPointer();
  Add32(FLUSHA);
  Add32(FLUSH);
  Add32(MSKPATH3(0));
  for(vifcode=0;vifcode<(24*4);vifcode++) Add32(NOP);
  Add32(MSKPATH3(0x8000));
  DMARet();
}

void CVIFStaticDMA::Initialise(int iNumPages)
{
	m_iNumPages = iNumPages;

	m_pPages = new CDMAMem[iNumPages];

	for(int iPage = 0; iPage < iNumPages; iPage++)
		SPS2Manager.Allocate(m_pPages[iPage], 256);

	*EE_VIF1_ERR = 2; // Ignore DMA mismatch error

	Begin();

	m_iPath3SyncCodes = VIFStaticDMA.GetPointer();

	Add32(FLUSHA);
	Add32(FLUSH);
	Add32(MSKPATH3(0));
	for(int vifcode = 0; vifcode < (24 * 4); vifcode++)
		Add32(NOP);
	Add32(MSKPATH3(0x8000));
	DMARet();
}

int CVIFStaticDMA::NewPage() {
        fprintf(stderr,"CVIFStaticDMA: %d/%d\n",m_iCurrPage,m_iNumPages);
	assert(m_iCurrPage < m_iNumPages);

	m_pPtr = (int *)m_pPages[m_iCurrPage].GetCached();

	m_pEndPtr = m_pPtr + PAGE_SIZE;
	int iPhysAddr = (int)m_pPages[m_iCurrPage++].GetPhysicalAddr();	
	m_iPhysicalOffset = iPhysAddr - (int)m_pPtr;

	return iPhysAddr;
}

CGIFDynamicDMA::CGIFDynamicDMA()
{
	m_iCurrPage = 0;
	m_iCurrBuffer = 0;
}

CGIFDynamicDMA::~CGIFDynamicDMA()
{

}

void CGIFDynamicDMA::Initialise(int iNumPages)
{
	m_iNumPages = iNumPages;

	// We are using a double buffer so allocate 2x the memory
	m_pPages = new CDMAMem[m_iNumPages * 2];

	for(int iPage = 0; iPage < (m_iNumPages * 2); iPage++)
		SPS2Manager.Allocate(m_pPages[iPage], 256);

	Begin();
}

int CGIFDynamicDMA::NewPage() {
	assert(m_iCurrPage < m_iNumPages);

	m_pPtr = (int *)m_pPages[m_iCurrPage + (m_iCurrBuffer * m_iNumPages)].GetCached();
	m_pEndPtr = m_pPtr + PAGE_SIZE;
	int iPhysAddr = (int)m_pPages[m_iCurrPage + (m_iCurrBuffer * m_iNumPages)].GetPhysicalAddr();
	++m_iCurrPage;

	return iPhysAddr;
}

void CGIFDynamicDMA::PrintPacket()
{
	SPS2Manager.FlushCache();

	unsigned int * pMem = (unsigned int *)m_pPages[(m_iCurrBuffer * m_iNumPages)].GetCached();
	
	printf("            ___________________________________________ \n");
	printf("  Address  |127     96|95      64|63      32|31       0| Virt Addr\n");
	printf(".----------+----------+----------+----------+----------|----------.\n");

	int iWritten = ((int)m_pPtr - (int)pMem + 0xF) >> 4;

	for(int i = 0; i < iWritten; i++)
	{
		static int iPhysAddr = 0;

		if((i % 256) == 0)
		{
			iPhysAddr = m_pPages[(i / 256) + (m_iCurrBuffer * m_iNumPages)].GetPhysicalAddr();
		}

		printf("| %.8X | ", iPhysAddr + (i % 256));
	
		for(int j = 0; j < 4; j++)
		{
			printf("%.8X | ", pMem[3 - j]);
		}

		printf("%.8X | ", (int)pMem);

		printf("\n");
		pMem += 4;
	}

	printf("'-----------------------------------------------------------------'\n");
}

void CGIFDynamicDMA::Fire(bool bFire = true, bool bPrint = false) {
int count;

	DMAEnd();

	if(bPrint)
		PrintPacket();

	asm("sync.l");
	SPS2Manager.FlushCache();
	
        if(bPrint) fprintf(stderr,"After FlushCache\n");

        count=0;

	while(*EE_D2_CHCR & 256) {
          count++;
        }

        if(bPrint) fprintf(stderr,"%d After while()\n",count);

	*EE_D2_QWC = 0;
	*EE_D2_TADR = m_iPhysAddr & 0x8ffffff0;

	if(bFire)
		*EE_D2_CHCR = 0x104;

	asm("sync.l");
	m_iCurrPage = 0;
	m_iCurrBuffer ^= 1;

	Begin();
}

CGIFStaticDMA::CGIFStaticDMA()
{
	m_iCurrPage = 0;
}

CGIFStaticDMA::~CGIFStaticDMA()
{

}

void CGIFStaticDMA::redo_from_start() {
  m_iCurrPage=0;
  Begin();
}

void CGIFStaticDMA::Initialise(int iNumPages)
{
	m_iNumPages = iNumPages;

	m_pPages = new CDMAMem[iNumPages];

	for(int iPage = 0; iPage < iNumPages; iPage++)
		SPS2Manager.Allocate(m_pPages[iPage], 256);

	Begin();
}

int CGIFStaticDMA::NewPage() {
        fprintf(stderr,"CGIFStaticDMA: %d/%d\n",m_iCurrPage,m_iNumPages);

	assert(m_iCurrPage < m_iNumPages);

	m_pPtr = (int *)m_pPages[m_iCurrPage].GetCached();

	m_pEndPtr = m_pPtr + PAGE_SIZE;
	int iPhysAddr = (int)m_pPages[m_iCurrPage++].GetPhysicalAddr();	
	m_iPhysicalOffset = iPhysAddr - (int)m_pPtr;

	return iPhysAddr;
}
