/* MAIN.CPP - Programmed by Jonathan Hobson */
/* 0404: Maarten Hofman: Changed to load the castle. The server connection.
   Debug what appears on screen. Allow the player to have a different name. Make
   certain models transparant. Made the camera controllable. Added font stuff.
   Made audio less demanding. Made a DEFINE to disable audio. Added asserts for
   the maxima. Fixed many problems and the elevator. Added keyboard support.
   Adding sound and vibration for error 997. Fixed placing of animation problem.
   Moved keyboard support. Changed hitpoint bar. */
/* 0504: Maarten Hofman: Added twod, fixed some scroll issues, added select
   mode, added the rotate command, added size to the who list and inventory,
   added "loading" window, added password, debugging an odd crash */
/* 0704: Maarten Hofman: Transparant models should be loaded last. Texture
   animation. Loading means loading, not doing funny stuff. Added performance
   timer. Added the examine/search option */
/* 0804: Maarten Hofman: Debugged font issue, added the second font, added
         relative moves, added fixed camera mode, added shop mode */
/* 270904: Maarten Hofman: Added the redo from start */
/* 251004: Maarten Hofman: Load fonts after redo from start */
/* 011104: Maarten Hofman: Added spell casting */
/* 081104: Maarten Hofman: Added more sound effects */
/* 301104: Maarten Hofman: Added displaying of texts */
/* 140205: Maarten Hofman: Added speed command */
/* 150205: Maarten Hofman: Fixed preloading of audio */
/* 160205: Maarten Hofman: Added auto refresh */
/* 230205: Maarten Hofman: Made reset dependent on slowness */
/* 280205: Maarten Hofman: Added light routines */

#include "config.h"

#include <stdio.h>
#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>
#include <memory.h>
#include <signal.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <strings.h>
#include <sys/time.h>

#include "ps2quaternion.h"
#include "ps2matrix4x4.h"
#include "ps2vector4.h"
#include "vumanager.h"
#include "sps2wrap.h"
#include "dma.h"
#include "pad.h"
#include "font.h"
#include "lists.h"
#include "keyboard.h"
#include "twod.h"
#include "selectmode.h"
#include "Ctimer.h"
#include "relative.h"
#include "music.h"
#include "light.h"

using namespace KazMaths;

char rcvbuf[100],sndbuf[300];
int cf;

// Pointers to the start and end of the VU code
extern "C" uint64 VU_vu1ms3dmodel_vu_start;
extern "C" uint64 VU_vu1ms3dmodel_vu_end;

// The main loop only runs as long as this flag is true
bool bLoop = true;

// Upon receiving an interrupt quit
void sig_handle(int sig) {
	bLoop = false;
}

// Simple function to alternate texture addresses.
// There are two texture areas, one at page 480 and
// one at page 496. Both are 16 pages in size
int TexArea(int & iTexArea)
{
	iTexArea ^= 1;
	return (iTexArea == 0) ? 496 : 480;
}

/* This function needs to be revised to send queued messages */
int send_all(int s,char *buf,int len) {
int total,bytesleft,n,ret;

  fprintf(stderr,"Send all: %c%c%c %d\n",buf[0],buf[1],buf[2],len);
  total=0;
  bytesleft=len;
  while(total<len) {
    n=send(s,buf+total,bytesleft,0);
    if(n==-1) break;
    total+=n;
    bytesleft-=n;
  }
  ret=n==-1?-1:0;
  fprintf(stderr,"Send all returned with: %d\n",ret);
  return ret;
}

int audiofd,sockfd,musicfd;

void open_path3(bool wait) {
const uint32 uVifCodesExit[] ALIGN_QW={ MSKPATH3(0x0),NOP,NOP,NOP,NOP};
  if(wait) {
    while(*EE_D1_CHCR & 256);
    while(*EE_VIF1_STAT&0x1f000003);
  }
  // Reopen path 3 before we exit
  DPUT_EE_VIF1_FIFO(((uint128 *) uVifCodesExit)[0]);
}

void close_path3(bool wait) {
const uint32 uVifCodesExit[] ALIGN_QW={ MSKPATH3(0x8000),NOP,NOP,NOP,NOP};
  if(wait) {
    while(*EE_D1_CHCR & 256);
    while(*EE_VIF1_STAT&0x1f000003);
  }
  // Reopen path 3 before we exit
  DPUT_EE_VIF1_FIFO(((uint128 *) uVifCodesExit)[0]);
}

void real_exit(int how) {

  close_keyboard_device();
  open_path3(false);
  pad_cleanup(PAD_0);
#ifdef AUDIO_ENABLED
  close_audio_device(audiofd);
  close_music_device(musicfd);
#endif
  close(sockfd);
  exit(how);
}

char *get_line(int sockfd) {
int ret,loc;

  loc=0;
  while(1) {
    ret=recv(sockfd,&rcvbuf[loc],1,0);
    if((ret==-1)&&(errno==EWOULDBLOCK)) return NULL;
    if(ret!=1) {
      perror("receiving data");
      real_exit(7);
    }
    if(loc==0) fprintf(stderr,"Getting line...\n");
    if(rcvbuf[loc]=='\n') break;
    loc++;
    if(loc>80) {
      fprintf(stderr,"No newline in response.\n");
      break;
    }
  }
  rcvbuf[loc]=0;
  fprintf(stderr,"Received: %s\n",rcvbuf);
  return rcvbuf;
}

bool zoom_towards(Vector3 *vCamEye,Vector3 *vCamAt,Vector3 *newCamEye,
                  Vector3 *newCamAt) {
bool moved;

  moved=false;
  if((newCamEye->x)-1>(vCamEye->x)) {
    vCamEye->x++;
    moved=true;
  }
  if((newCamEye->x)+1<(vCamEye->x)) {
    vCamEye->x--;
    moved=true;
  }
  if((newCamEye->y)-1>(vCamEye->y)) {
    vCamEye->y++;
    moved=true;
  }
  if((newCamEye->y)+1<(vCamEye->y)) {
    vCamEye->y--;
    moved=true;
  }
  if((newCamEye->z)-1>(vCamEye->z)) {
    vCamEye->z++;
    moved=true;
  }
  if((newCamEye->z)+1<(vCamEye->z)) {
    vCamEye->z--;
    moved=true;
  }
  if((newCamAt->x)-1>(vCamAt->x)) {
    vCamAt->x++;
    moved=true;
  }
  if((newCamAt->x)+1<(vCamAt->x)) {
    vCamAt->x--;
    moved=true;
  }
  if((newCamAt->y)-1>(vCamAt->y)) {
    vCamAt->y++;
    moved=true;
  }
  if((newCamAt->y)+1<(vCamAt->y)) {
    vCamAt->y--;
    moved=true;
  }
  if((newCamAt->z)-1>(vCamAt->z)) {
    vCamAt->z++;
    moved=true;
  }
  if((newCamAt->z)+1<(vCamAt->z)) {
    vCamAt->z--;
    moved=true;
  }
  return moved;
}

#ifdef AUDIO_ENABLED
soundbyte *nowsound;

void handle_audio() {
unsigned char *audiobuf;
int ff,i;

  if(!nowsound) return;
  /* Handle audio */
  ff=check_free_fragments(audiofd);
  if(!ff) return;
  for(i=0;i<ff;i++) {
    audiobuf=nowsound->get_next_fragment();
    if(!audiobuf) audiobuf=silence;
    write(audiofd,audiobuf,frag_size);
  }
  if(audiobuf==silence) {
    nowsound=nowsound->get_link();
  }
}

void handle_music() {
int ff,i;

  ff=check_free_fragments(musicfd);
  if(!ff) return;
  for(i=0;i<ff;i++) fill_music_buffer(musicfd);
}

void play_sound(char *which) {
int i;

  for(i=0;i<sm;i++) {
    if(sounds[i]->has_name(which)) {
      /* No need to play the current sound again */
      if(sounds[i]==nowsound) return;
      if(nowsound) {
        nowsound->set_link(sounds[i]);
        sounds[i]->reset();
        fprintf(stderr,"Queueing sound %s\n",which);
        return;
      }
      nowsound=sounds[i];
      nowsound->reset();
      fprintf(stderr,"Playing sound %s\n",which);
      return;
    }
  }
  fprintf(stderr,"Sound %s not found!\n",which);
}
#endif

int main(int argc, char **argv) {
struct hostent *he;
struct sockaddr_in servd_addr;
int ret,i,j,k,value,tx,ty,tz,tick,dl,hp,newhp,maxhp,vibrate,padmode,cc,count;
int exastate,dir,redostate,sp,maxsp,speed,ccreset,tune,fpslowcnt,fpscur;
int newlight;
char *response,nm[32],mdl[32],txtr[32],*action,*my_name,trigger[100],text[100];
char *saction;
Matrix4x4 *matWorld,matView;
float cx,cy,cz,angle,opa;
bool is_moving,is_loading,do_loop,recalc_camera;
CFont font,addfont;
twod twod;
CTexture fontex,addfontex,*animate,dummy;
blob *current,*player;
struct blobcopy bc;
window *text_output,*square_warning,*text_input;

        /* Check for sufficient commandline parameters */
        if(argc<5) {
          fprintf(stderr,
                  "Usage: KazEngine <server> <port> <name> <password>\n");
          exit(1);
        }
        action=NULL;
        trigger[0]=0;
        text[0]=0;
        vibrate=0;
        is_moving=false;
        is_loading=false;
        exastate=0;
        cc=0;
        dl=2;
        cf=0;
        redostate=0;
        shopmode=false;
        recalc_camera=false;
        player=NULL;
        fpslowcnt=0;

#ifdef AUDIO_ENABLED
        /* Initialise audio */
        audiofd=open_audio_device();
        musicfd=open_music_device();
        /* Set opening music */
        tune=-1;
#endif

        hp=0;
        maxhp=0;
        sp=1;
        maxsp=1;
        /* Initialise the connection to the server */
        he=gethostbyname(argv[1]);
        if(he==NULL) {
          perror("gethostbyname");
          exit(2);
        }
        servd_addr.sin_family=AF_INET;
        servd_addr.sin_port=htons(atoi(argv[2]));
        servd_addr.sin_addr=*((struct in_addr *)he->h_addr);
        bzero(&(servd_addr.sin_zero),8);
        if((sockfd=socket(AF_INET,SOCK_STREAM,0))==-1) {
          perror("server socket");
          exit(3);
        }
        ret=connect(sockfd,(struct sockaddr *)&servd_addr,
                    sizeof(struct sockaddr));
        if(ret==-1) {
          perror("server connect");
          exit(4);
        }
        fcntl(sockfd,F_SETFL,O_NONBLOCK);
        my_name=argv[3];
        /* argv[4] is the password */
        sprintf(sndbuf,"login %s %s\n",my_name,argv[4]);
        ret=send_all(sockfd,sndbuf,strlen(sndbuf));
        if(ret==-1) {
          perror("server send");
          exit(5);
        }

	if(!pad_init(PAD_0, PAD_INIT_LOCK | PAD_INIT_ANALOGUE | PAD_INIT_PRESSURE))
	{
		fprintf(stderr,"Failed to initialise control pad\n");
		pad_cleanup(PAD_0);
		exit(-1);
	}
	
        /* Check for actuator */
        if(pad[0].actuator) {
          enable_actuator(0,1,1);
        }
        else {
          vibrate=-1;
        }

        /* Switch to basic moving around mode */
        padmode=0;

	SPS2Manager.Initialise(2048);	// Allocate 2048 pages (8MB)

	for (int sig=0; sig<128; sig++)
		signal(sig, sig_handle);

	// Allocate 2MB for static VIF packets
	VIFStaticDMA.Initialise(512);
	// Allocate 1MB for dynamic VIF packets (double buffered)
	VIFDynamicDMA.Initialise(256);

	// Allocate 2MB for static GIF packets (mostly textures)
	GIFStaticDMA.Initialise(512);
	// Allocate 1MB for dynamic GIF packets (double buffered)
	GIFDynamicDMA.Initialise(256);

	// Create a right handed projection matrix.
	float n = 0.1f;		// Near plane
	float f = 1000.0f;	// Far plane
	float FovYdiv2 = DegToRad(60.0f) * 0.5f; // 60 degree vertical FOV
	float cotFOV = 1.0f / (Sin(FovYdiv2) / Cos(FovYdiv2));
	float w = cotFOV * (SPS2Manager.GetScreenWidth() / 4096.0f) / (4.0f / 3.0f);
	float h = cotFOV * (SPS2Manager.GetScreenHeight() / 4096.0f);

	Matrix4x4 matProj
		(w, 0, 0, 0,
		 0, -h, 0, 0,
		 0, 0, (f+n) / (f-n), -1,
		 0, 0, (2*f*n) / (f-n), 0);

	// Upload the VU1 code for rendering the Milkshape models
	CVU1MicroProgram VU1Ms3dModel(&VU_vu1ms3dmodel_vu_start, &VU_vu1ms3dmodel_vu_end);
	VU1Ms3dModel.Upload();

	Vector3 vCamEye(0,35.0f,80.0f),vCamAt(0,35.0f,0);
        Vector3 oldCamEye,newCamEye,newCamAt,oldCamAt;

	int iTexArea = 0;	// Start with texture buffer 0

        mm=0;
        tm=0; 
        bm=0;
        tick=0;
        setup_light();
        set_light(1);
        fprintf(stderr,"Waiting for server response...\n");
        while(1) {
          if(tick++>100000) {
            fprintf(stderr,"Failed to receive response.\n");
            real_exit(8);
          }
          response=get_line(sockfd);
          if(response==NULL) continue;	// Still loading
          value=atoi(response);
          if(!value) continue;		// Error!
          if(value==5) break;		// Description end
          if(value==9) {		// Biasing of shadows
            sscanf(response,"009 BIAS %s",nm);
            current=find_blob(nm);
            if(current) current->set_bias();
          }
#ifdef AUDIO_ENABLED
          if(value==11) {		// Preloading audio
            sscanf(response,"011 AUDIO %s",nm);
            load_sound(nm);
          }
#endif
          if(value==13) {		// Animations
            sscanf(response,"013 ANIMATE %s",nm);
            current=find_blob(nm);
            if(current) current->set_animate();
          }
          if(value==17) {		// Rotation animations
            if(sscanf(response,"017 ROTATEY %s",nm)==1) {
              current=find_blob(nm);
              if(current) current->set_rotate(2);
            }
            else {
              sscanf(response,"017 ROTATE %s X %f Y %f Z %f",nm,&cx,&cy,&cz);
              current=find_blob(nm);
              if(current) current->do_rotate(cx,cy,cz,1);
            }
          }
          if(value==23) {
            sscanf(response,"023 MUSIC %i",&tune);
            if(tune>=0) set_tune(tune);
          }
          if(value==24) {
            sscanf(response,"024 LIGHT %i",&newlight);
            set_light(newlight);
          }
          if(value==12) sscanf(response,"012 HP %d MAX %d",&hp,&maxhp);
          if(value==20) sscanf(response,"020 SP %d MAX %d",&sp,&maxsp);
          if(value==3) {		// Position model
            if(sscanf(response,"003 LOCATE %s X %d Y %d Z %d",nm,&tx,&ty,
                      &tz)!=4) {
              fprintf(stderr,"Incorrect sscanf\n");
            }
            fprintf(stderr,"Looking for %s translate to <%d,%d,%d>\n",nm,tx,ty,
                    tz);
            for(i=0;i<bm;i++) {
              if(blobs[i]->has_name(nm)) {
                fprintf(stderr,"Name found %d\n",i);
                if(blobs[i]->move_coordinates(tx,ty,tz)) {
                  if(!strcmp(nm,my_name)) {
                    player=blobs[i];
                    fprintf(stderr,"Player moved.\n");
                    /* Automagically move camera to above the player */
                    if(check_camera(&vCamEye,tx,ty,tz)) {
                      vCamAt.x=tx;
                      vCamAt.y=ty;
                      vCamAt.z=tz;
                      recalc_camera=true;
                    }
                  }
                }
              }
              else {
                if(blobs[i]->has_base_name(nm)) blobs[i]->update();
              }
            }
            continue;
          }
          if(value==1) {		// Load model and texture
            if(sscanf(response,"001 NAME %s MODEL %s TEXTURE %s",nm,mdl,
               txtr)!=3) {
              fprintf(stderr,"Incorrect sscanf\nN: %s Model: %s Texture: %s\n",
                      nm,mdl,txtr);
            }
            else {
              load_model_and_texture(nm,mdl,txtr);
            }
          }
          if(value==18) {		// Conditional update blob
            sscanf(response,"018 CUPDATE %s MODEL %s TEXTURE %s",nm,mdl,txtr);
            current=find_blob(nm);
            if(current) {
              current->set_copy(&bc);
              unload_blob(nm);
              load_model_and_texture(nm,mdl,txtr);
              current=find_blob(nm);
              if((current)&&(bc.parent!=NULL)) {
                current->set_parent(bc.parent);
                current->set_relative(bc.ax,bc.ay,bc.az);
                current->set_animation_type(bc.animtype,bc.animspeed);
                current->set_animation(bc.animation);
                current->set_updated();
                current->set_angle(bc.angle);
                current->update();
              }
            }
          }
        }
	
	// Load the font width data
        sprintf(sndbuf,"%s%s",RESOURCE_DIR,"font.dat");
	font.Load(sndbuf);
        fprintf(stderr,"Loading font texture...\n");
        sprintf(sndbuf,"%s%s",RESOURCE_DIR,"font_.tm2");
        fontex.LoadImage(sndbuf,true);
        sprintf(sndbuf,"%s%s",RESOURCE_DIR,"addfont8_.tm2");
        addfontex.LoadImage(sndbuf);
        sprintf(sndbuf,"%s%s",RESOURCE_DIR,"waterc1_.tm2");
        dummy.LoadImage(sndbuf);

        open_keyboard_device();

        /* Load all sounds, to get rid of the preloading issues */
        load_all_sounds();

        text_output=twod.open_window(-300,150,300,200);
        square_warning=twod.open_window(-100,-200,100,-175);

        CTimer timer;
#ifdef TIMER_ENABLED
        window *fps_display;

        fps_display=twod.open_window(200,-200,300,-175);
#endif
        text_input=twod.open_window(-300,-200,-125,-175);
        initialize_select_mode(&twod);
        fprintf(stderr,"mm: %d, tm: %d, bm: %d\n",mm,tm,bm);

	while(bLoop) {
#ifdef TIMER_ENABLED
          timer.PrimeTimer();
#endif

          // Fire the packets built in the previous frame
          if(redostate!=1) {
            GIFDynamicDMA.Fire();	
	    VIFDynamicDMA.Fire();
            close_path3(true);
          }
          if(vibrate>0) {
            vibrate--;
            if(!vibrate) set_actuator(0,0,0);
          }

          if(cc>0) {
            cc++;
            if(cc==10) cc=0;
          }

          /* Camera updates should always take place */
	  pad_update(PAD_0);
          cx=0;
          cz=0;
          cx-=pad[0].axes[PAD_AXIS_RX];
          cz+=pad[0].axes[PAD_AXIS_RY];
          if(exastate==0) {
            if((cx>0.1)||(cx<-0.1)||(cz>0.1)||(cz<-0.1)) {
              vCamEye.x+=cx;
              vCamEye.z+=cz;
              recalc_camera=true;
            }
#ifdef FIXED_CAMERA
            if((pad[0].buttons & PAD_R3)&&(player!=NULL)) {
              angle=player->get_angle();
              cf=0;
              if(angle<3) cf++;
              if(angle<2) cf++;
              if(angle<1.5) cf++;
              if(angle<0.5) cf++;
              if(angle<-0.5) cf++;
              if(angle<-1.5) cf++;
              if(angle<-2) cf++;
              switch(cf) {
                case 0:
                  cf=4;
                  break;
                case 1:
                  cf=3;
                  break;
                case 3:
                  cf=1;
                  break;
                case 4:
                  cf=0;
                  break;
                case 5:
                  cf=7;
                  break;
                case 7:
                  cf=5;
              }
              recalc_camera=check_camera(&vCamEye,vCamAt.x,vCamAt.y,vCamAt.z);
            }
#endif
          }
          if(dl) fprintf(stderr,"Start of pad checks\n");

          ccreset=0;
          if((pad[0].buttons & PAD_R2) && (pad[0].buttons & PAD_L2)) ccreset=1;
          if(fpslowcnt>20) {
            fpslowcnt=0;
            ccreset=1;
          }
          if(ccreset) {
            if(redostate==0) {
              redostate=1;
              opa=10;
              if(player) opa=player->get_angle();
              send_all(sockfd,"look\n",5);
              VIFStaticDMA.redo_from_start();
              GIFStaticDMA.redo_from_start();
              redo_lists_from_start();
              setup_light();
              set_light(1);
              /* Also reload the fonts and other preloaded stuff */
              sprintf(sndbuf,"%s%s",RESOURCE_DIR,"font_.tm2");
              fontex.LoadImage(sndbuf,true);
              sprintf(sndbuf,"%s%s",RESOURCE_DIR,"addfont8_.tm2");
              addfontex.LoadImage(sndbuf);
              sprintf(sndbuf,"%s%s",RESOURCE_DIR,"waterc1_.tm2");
              dummy.LoadImage(sndbuf);
            }
          }
          else {
            if(redostate==2) redostate=0;
          }

          if(exastate==0) {
            if((pad[0].buttons & PAD_L1)&&(padmode<2)) {
              exastate=1;
              send_all(sockfd,"examine\n",8);
              oldCamEye.x=vCamEye.x;
              oldCamEye.y=vCamEye.y;
              oldCamEye.z=vCamEye.z;
              oldCamAt.x=vCamAt.x;
              oldCamAt.y=vCamAt.y;
              oldCamAt.z=vCamAt.z;
            }
          }
          else {
            if(!(pad[0].buttons & PAD_L1)) {
              exastate=0;
              vCamEye.x=oldCamEye.x;
              vCamEye.y=oldCamEye.y;
              vCamEye.z=oldCamEye.z;
              vCamAt.x=oldCamAt.x;
              vCamAt.y=oldCamAt.y;
              vCamAt.z=oldCamAt.z;
              recalc_camera=true;
            }
            else {
              if(exastate==2) {
                /* Move camera closer to point */
                if(zoom_towards(&vCamEye,&vCamAt,&newCamEye,&newCamAt)) {
                  recalc_camera=true;
                }
              }
            }
          }

          if((!is_loading)&&(cc==0)&&(exastate==0)) {
		if((pad[0].buttons & PAD_START)&&(pad[0].buttons&PAD_SELECT)) {
		  bLoop = false;
                }

                if((!padmode)&&(pad[0].buttons&PAD_SELECT)) padmode=1;
                if((padmode==1)&&(!(pad[0].buttons&PAD_SELECT))) {
                  start_select_mode(0);
                  padmode=2;
                }
                if((padmode==2)&&(pad[0].buttons&PAD_SELECT)) padmode=3;
                if((padmode==3)&&(!(pad[0].buttons&PAD_SELECT))) {
                  end_select_mode();
                  padmode=0;
                }

		// Create and fill the axis variables
		float x = 0.0f, y = 0.0f;

		x += pad[0].pressures[PAD_PRIGHT];
		x -= pad[0].pressures[PAD_PLEFT];
		y += pad[0].pressures[PAD_PUP];
		y -= pad[0].pressures[PAD_PDOWN];
		x += pad[0].axes[PAD_AXIS_LX];
		y -= pad[0].axes[PAD_AXIS_LY];

                if(padmode<2) {
                  /* Normal game mode inputs */
                  if((action==NULL)&&(!is_moving)) {
                    dir=calculate_direction(x,y);
                    if(dir>=0) {
#ifdef RELATIVE_MOVES
                      dir+=calculate_segment(vCamAt.x,vCamAt.z,vCamEye.x,
                                             vCamEye.z);
#endif
                      action=translate_direction(dir);
                    }
                    if(trigger[0]) {
                      if(pad[0].buttons & PAD_SQUARE) action="trigger\n";
                    }
                    if(action!=NULL) {
                      if(action[0]!='t') {
                        if(pad[0].buttons & PAD_CROSS) {
                          sprintf(sndbuf,"attack %s",action);
                        }
                        else {
                          /* The CIRCLE button is used to cast spells */
                          if(pad[0].buttons & PAD_CIRCLE) {
                            sprintf(sndbuf,"spell %s",action);
                          }
                          else {
                            sprintf(sndbuf,"move %s",action);
                          }
                        }
                        action=sndbuf;
                      }
                      fprintf(stderr,"%s",action);
                      ret=send_all(sockfd,action,strlen(action));
                      cc=1;
                      if(ret<0) {
                        perror("ingame send_all");
                        action=NULL;
                      }
                    }
                  }
                }
                else {
                  saction=handle_select_mode(x,y);
                  if(saction!=NULL) send_all(sockfd,saction,strlen(saction));
                }
        }
        if(dl) fprintf(stderr,"Start of server response\n");
        do_loop=true;
        /* Check if there is a server response, and handle that line */
        while((do_loop)&&((response=get_line(sockfd))!=NULL)) {
                  action=NULL;
                  fprintf(stderr,"Server message: %s\n",response);
                  value=atoi(response);
#ifdef AUDIO_ENABLED
                  if(value==10) {		// Play a sound
                    sscanf(response,"010 %s",nm);
                    play_sound(nm);
                  }
                  if(value==997) {		// Hit someone/something
                    if(!vibrate) {
                      vibrate=10;
                      if(option_vibrate) set_actuator(0,1,255);
                    }
                    play_sound("nono");
                  }
                  if(value==23) {
                    sscanf(response,"023 MUSIC %i",&tune);
                    if(tune>=0) set_tune(tune);
                  } 
                  if(value==24) {
                    sscanf(response,"024 LIGHT %i",&newlight);
                    set_light(newlight);
                  }
                  if(value==11) {		// Preload audio
                    sscanf(response,"011 AUDIO %s",nm);
                    load_sound(nm);
                  }
#endif
                  if(value==5) {		// Description end
                    is_loading=false;
                    if(redostate==1) redostate=2;
                    dl=1;
                  }
                  if(value==13) {		// Animate a blob
                    sscanf(response,"013 ANIMATE %s",nm);
                    current=find_blob(nm);
                    if(current) current->set_animate();
                  }
                  if(value==22) {		// Speed blob
                    sscanf(response,"022 SPEED %s %i",nm,&speed);
                    current=find_blob(nm);
                    if(current) current->set_move_speed(speed);
                  }
                  if(value==21) {
                    start_select_mode(1);
                    padmode=2;
                  }
                  if((value==17)||(value==19)) {	// Rotation animations
                    if(sscanf(response,"017 ROTATEY %s",nm)==1) {
                      current=find_blob(nm);
                      if(current) current->set_rotate(2);
                    }
                    else {
                      sscanf(response,"0%*d ROTATE %s X %f Y %f Z %f",nm,&cx,
                             &cy,&cz);
                      current=find_blob(nm);
                      if(current) current->do_rotate(cx,cy,cz,value==17);
                    }
                  }
                  if(value==20) sscanf(response,"020 SP %d MAX %d",&sp,&maxsp);
                  if(value==12) {		// Update HP and MAXHP
                    sscanf(response,"012 HP %d MAX %d",&newhp,&maxhp);
                    if(newhp<hp) {
                      play_sound("pain");
                      if(!vibrate) {
                        vibrate=10;
                        if(option_vibrate) set_actuator(0,1,255);
                      }
                    }
                    hp=newhp;
                  }
                  if(value==9) {
                    sscanf(response,"009 BIAS %s",nm);
                    current=find_blob(nm);
                    if(current) current->set_bias();
                  }
                  if(value==3) {		// Position blob
                    if(sscanf(response,"003 LOCATE %s X %d Y %d Z %d",nm,&tx,
                       &ty,&tz)!=4) {
                      fprintf(stderr,"Incorrect sscanf\n");
                    }
                    for(i=0;i<bm;i++) {
                      if(blobs[i]->has_name(nm)) {
                        fprintf(stderr,"Name found %d\n",i);
                        if(blobs[i]->move_coordinates(tx,ty,tz)) {
                          if(!strcmp(nm,my_name)) {
                            player=blobs[i];
                            if(redostate==1) {
                              if(opa<9) {
                                player->set_angle(opa);
                                fprintf(stderr,"Setting angle %f.\n",opa);
                              }
                            } 
                            fprintf(stderr,"Player moved.\n");
                            /* Automagically move camera to above the player */
                            if(check_camera(&vCamEye,tx,ty,tz)) {
                              vCamAt.x=tx;
                              vCamAt.y=ty;
                              vCamAt.z=tz;
                              recalc_camera=true;
                            }
                          }
                        }
                      }
                      else {
                        if(blobs[i]->has_base_name(nm)) blobs[i]->update();
                      }
                    }
                  }
                  if(value==1) {		// Load model and texture
                    if(sscanf(response,"001 NAME %s MODEL %s TEXTURE %s",nm,mdl,
                              txtr)==3) {
                      load_model_and_texture(nm,mdl,txtr);
                    }
                  }
                  if(value==4) {		// Camera positioning
                    if(exastate==1) {
                      sscanf(response,"004 CAMERA EYE %f,%f,%f AT %f,%f,%f",
                        &(newCamEye.x),&(newCamEye.y),&(newCamEye.z),
                        &(newCamAt.x),&(newCamAt.y),&(newCamAt.z));
                      exastate=2;
                    }
                  }
                  if(value==18) {		// Conditional update blob
                    sscanf(response,"018 CUPDATE %s MODEL %s TEXTURE %s",nm,mdl,
                           txtr);
                    current=find_blob(nm);
                    if(current) {
                      /* Only update conditionally */
                      if(!current->get_updated()) {
                        current->set_copy(&bc);
                        unload_blob(nm);
                        load_model_and_texture(nm,mdl,txtr);
                        current=find_blob(nm);
                        if((current)&&(bc.parent!=NULL)) {
                          current->set_parent(bc.parent);
                          current->set_relative(bc.ax,bc.ay,bc.az);
                          current->set_animation_type(bc.animtype,bc.animspeed);
                          current->set_animation(bc.animation);
                          current->set_updated();
                          current->update();
                        }
                      }
                      else {
                        fprintf(stderr,"Blob %s already updated.\n",nm);
                      }
                    }
                  }
                  if(value==14) {		// Update blob with model
                    if(sscanf(response,"014 UPDATE %s MODEL %s TEXTURE %s",nm,
                              mdl,txtr)==3) {
                      current=find_blob(nm);
                      if(current) {
                        current->set_copy(&bc);
                        unload_blob(nm);
                        load_model_and_texture(nm,mdl,txtr);
                        current=find_blob(nm);
                        if((current)&&(bc.parent!=NULL)) {
                          current->set_parent(bc.parent);
                          current->set_relative(bc.ax,bc.ay,bc.az);
                          current->set_animation_type(bc.animtype,bc.animspeed);
                          current->set_animation(bc.animation);
                          current->set_updated();
                          current->update();
                        }
                      }
                    }
                  }
                  if(value==15) {		// Remove blob completely
                    if(sscanf(response,"015 REMOVE %s",nm)==1) {
                      unload_blob(nm);
                    }
                  }
                  if(value==6) {		// Remove trigger
                    trigger[0]=0;
                    square_warning->clear();
                    shopmode=false;
                  }
                  if(value==7) {		// Add trigger
                    /* Check for special "shop" trigger */
                    if(strcmp(response,"007 shop")==0) {
                      shopmode=true;
                      square_warning->clear();
                      square_warning->add_line("SELECT Shop available");
                    }
                    else {
                      sprintf(trigger,"   %s",&response[4]);
                      square_warning->clear();
                      square_warning->add_line(trigger);
                      square_warning->add_symbol(1,1,4);
                    }
                  }
                  if(value==8) {		// Add message text
                    strncpy(text,&response[4],100);
                    text[99]=0;
                    text_output->add_line(text); 
                  }
                  if(value==16) {		// Who list/inventory input
                    select_response(response);
                  }

                  /* Check for loading after the loading check */
                  if(value==2) {		// Description start
                    is_loading=true;
                    /* We encountered a new location */
                    dl=255;
                  }

                  /* No repeated checks if we're not loading */
                  if(!is_loading) do_loop=false;
        }

        /* No need to do anything if we're in redostate */
        if(redostate==1) continue;

        if(dl) fprintf(stderr,"Start of handle keyboard\n");
        if(handle_keyboard(text_input)) {
          /* We received a line from the keyboard */
          fprintf(stderr,"Keyboard: %s\n",keybuf);
          if(pad[0].buttons & PAD_CIRCLE) {
            sprintf(sndbuf,"shout %s",keybuf);
          }
          else {
            sprintf(sndbuf,"say %s",keybuf);
          }
          if(send_all(sockfd,sndbuf,strlen(sndbuf))<0) {
            perror("Send all.\n");
          }
          /* Read, so reset keyboard */
          keybuf[0]=0;
          kbc=0;
          text_input->clear();
        }

        /* Run the textures */
        for(i=0;i<tm;i++) {
          animate=texs[i]->get_next();
          if(animate!=NULL) {
            if(dl) fprintf(stderr,"Running animated texture %d\n",i);
            texs[i]=animate;
          }
        }
        if(dl) fprintf(stderr,"Start of texture upload\n");
        /* First upload opaque textures */
        for(i=0;i<tm;i++) {
          if(!(texs[i]->is_trans())) texs[i]->Upload(TexArea(iTexArea));
        }
        /* Then the transparant ones */
        for(i=0;i<tm;i++) {
          if(texs[i]->is_trans()) texs[i]->Upload(TexArea(iTexArea));
        }
        fontex.Upload(TexArea(iTexArea));
        addfontex.Upload(TexArea(iTexArea));
        dummy.Upload(TexArea(iTexArea));

        if(dl) fprintf(stderr,"First open Path 3\n");
        VIFDynamicDMA.OpenPath3();
        if(dl) fprintf(stderr,"BeginScene()\n");
	SPS2Manager.BeginScene(0, 0, 30);	// Clear the screen to a dark blue
        if(dl) fprintf(stderr,"Second open Path 3\n");
        VIFDynamicDMA.OpenPath3();
	/* Run the blobs */
        is_moving=false;
        if(dl) fprintf(stderr,"Running the blobs\n");
        for(i=0;i<bm;i++) {
                  if(blobs[i]->heart_beat()) {
                    if(blobs[i]->has_name(my_name)) {
                      /* 
                       * Automagically move camera to above the player if the
                       * player isn't in examine mode.
                       */
                      player=blobs[i];
                      if(exastate==0) {
                        check_camera(&vCamEye,blobs[i]->get_cx(),
                                     blobs[i]->get_cy(),blobs[i]->get_cz());
                        vCamAt.x=blobs[i]->get_cx();
                        vCamAt.y=blobs[i]->get_cy();
                        vCamAt.z=blobs[i]->get_cz();
	                recalc_camera=true;
                        if(blobs[i]->target_distance()>0.5) is_moving=true;
                      }
                    }
                  }
        }

        if(recalc_camera) {
          matView.LookAt(vCamEye,vCamAt,Vector3(0,1,0));
          recalc_camera=false;
        }
        if(dl) fprintf(stderr,"Start of rendering the scene\n");
        /* First opaque, then transparant */
        count=0;
        for(k=0;k<2;k++) {
          for(i=0;i<tm;i++) {
            /* Separate the transparants from the opaques */
            if(!k) {
              if(texs[i]->is_trans()) continue;
            }
            else {
              if(!(texs[i]->is_trans())) continue;
            }
            if(dl) fprintf(stderr,"Selecting texture %d\n",i);
            texs[i]->Select();
            /* Is called once for each texture, and finally for the font */
            for(j=0;j<bm;j++) {
              if(blobs[j]->get_texture()!=i) continue;
              if(dl) fprintf(stderr,"Writing blob %d\n",j);
              matWorld=blobs[j]->get_location();
              models[blobs[j]->get_model()]->Render((*matWorld)*matView*matProj,*matWorld);
            }
            if(count++<=tm) {
              VIFDynamicDMA.OpenPath3();
              if(dl) fprintf(stderr,"Next OpenPath3\n");
            }
          }
        }

#ifdef AUDIO_ENABLED
        if(dl) fprintf(stderr,"Handling audio\n");
        if(option_sound) handle_audio();
        if(tune>=0) if(option_music) handle_music();
#endif

        if(dl) fprintf(stderr,"Generating hitpoint bar\n");
        twod.hitpoint_bar(-300,-210,hp,maxhp);
        fpscur=(int)(timer.GetFPS());
#ifdef TIMER_ENABLED
        fps_display->clear();
        sprintf(sndbuf,"%d",fpscur);
        fps_display->add_line(sndbuf);
#endif
        if(fpscur<60) {
          fpslowcnt++;
        }
        else {
          fpslowcnt=0;
        }

        /* Now select the font, and write necessary texts */
        if(dl) fprintf(stderr,"Writing windows\n");
        fontex.Select();
        if(count++<=tm) {
          VIFDynamicDMA.OpenPath3();
          if(dl) fprintf(stderr,"Next OpenPath3\n");
        }
        twod.render(&font,1);
        addfontex.Select();
        twod.render(&addfont,2);
        dummy.Select();
	SPS2Manager.EndScene();
        if(dl>0) {
          dl--;
          if(!dl) fprintf(stderr,"End of loop documentation.\n");
        }
  }
  real_exit(0);
}

