/* MUSIC.C - Programmed by Maarten Hofman */
/* 010904: Started programming */
/* 210205: Adapted for game */

#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <math.h>
#include <unistd.h>
#include "music.h"

#define VOICES	2

float step[VOICES],loc[VOICES],notes[12*7];
int tunebass[4],tunevoice[4],amp[VOICES],decay[VOICES],currentune;
int type[VOICES],beatc,placecount,tunespeed,measurenotes,octave,beattype;
char triangle[512],sawtooth[512],sine[512];

void initialise_types();

void set_tune(int which) {
  if(which==currentune) return;
  currentune=which;
  octave=36;
  beatc=10000;
  placecount=0;
  switch(which) {
    case 0:
      tunebass[0]=0;
      tunebass[1]=5;
      tunebass[2]=0;
      tunebass[3]=7;
      tunevoice[0]=4;
      tunevoice[1]=9;
      tunevoice[2]=11;
      tunevoice[3]=2;
      measurenotes=4;
      tunespeed=2000;
      beattype=1;
      return;
    case 2:
      tunebass[0]=0;
      tunebass[1]=2;
      tunebass[2]=4;
      tunebass[3]=5;
      tunevoice[0]=7;
      tunevoice[1]=9;
      tunevoice[2]=11;
      tunevoice[3]=0;
      measurenotes=3;
      tunespeed=2500;
      beattype=2;
      return;
    default:
      tunebass[0]=9;
      tunebass[1]=7;
      tunebass[2]=5;
      tunebass[3]=4;
      tunevoice[0]=0;
      tunevoice[1]=2;
      tunevoice[2]=11;
      tunevoice[3]=9;
      placecount=0;
      measurenotes=4;
      tunespeed=3000;
      beattype=0;
      return;
  }
}

int open_music_device() {
int music_fd,format,stereo,speed;

  format=AFMT_S8;
  stereo=0;
  speed=11025;
  music_fd=open("/dev/dsp1",O_WRONLY,0);
  if(music_fd==-1) {
    perror("/dev/dsp");
    exit(1);
  }
  if(ioctl(music_fd,SNDCTL_DSP_SETFMT,&format)==-1) {
    perror("SNDCTL_DSP_SETFMT");
    exit(2);
  }
  if(ioctl(music_fd,SNDCTL_DSP_STEREO,&stereo)==-1) {
    perror("SNDCTL_DSP_STEREO");
    exit(3);
  }
  initialise_types();
  return music_fd;
}

void close_music_device(int fd) {
  close(fd);
}

void set_parameters(int voice,float s,int a,int d,int t) {
  step[voice]=s;
  amp[voice]=a;
  decay[voice]=d;
  type[voice]=t;
  loc[voice]=0;
}

inline int get_next_byte(int which) {
int value;

  switch(type[which]) {
    case 0:
      value=triangle[(int)loc[which]];
      break;
    case 1:
      value=sawtooth[(int)loc[which]];
      break;
    case 2:
      value=sine[(int)loc[which]];
      break;
    case 3:
      value=(random()&(0xff))-128;
      break;
  }
  if(amp[which]<32767) {
    value>>=1;
    if(amp[which]>16382) value=value+value>>2;
  }
  if(amp[which]<16383) value>>=1;
  if(amp[which]<8191) value>>=1;
  if(amp[which]<4095) value>>=1;
  if(amp[which]<2047) value>>=1;
  if(amp[which]<1023) value>>=1;
  if(amp[which]<511) value>>=1;
  if(amp[which]<255) value>>=1;
  loc[which]+=step[which];
  if(loc[which]>511) loc[which]=loc[which]-511;
  amp[which]-=decay[which];
  if(amp[which]<0) amp[which]=0;
  return value;
}

void next_beat() {
float bass;
float note;
int rnd;

  placecount++;
  if(placecount>=measurenotes*4) placecount=0;
  switch(beattype) {
    case 0:
      if(!(placecount%measurenotes)) {
        bass=notes[tunebass[placecount/measurenotes]+12];
        set_parameters(1,bass,40000,4,1);
      }
      break;
    case 1:
      if(!(placecount%measurenotes)) {
        bass=notes[tunebass[placecount/measurenotes]];
        set_parameters(1,bass,40000,4,1);
      }
      if(!((placecount+2)%measurenotes)) {
        bass=notes[tunebass[placecount/measurenotes]+12];
        set_parameters(1,bass*2,40000,4,1);
      }
      break;
    case 2:
      if(!(placecount%measurenotes)) {
        set_parameters(1,1,40000,10,3);
      }
      break;
  }
  rnd=random()&15;
  printf("%d\n",rnd);
  switch(rnd) {
    case 0:
    case 11:
      return;
    case 1:
      octave+=12;
      if(octave>48) octave=48;
      return;
    case 2:
      octave-=12;
      if(octave<24) octave=24;
      return;
    case 3:
    case 12:
      note=notes[tunebass[0]+octave];
      break;
    case 4:
    case 13:
      note=notes[tunebass[1]+octave];
      break;
    case 5:
    case 14:
      note=notes[tunebass[2]+octave];
      break;
    case 6:
    case 15:
      note=notes[tunebass[3]+octave];
      break;
    case 7:
      note=notes[tunevoice[0]+octave];
      break;
    case 8:
      note=notes[tunevoice[1]+octave];
      break;
    case 9:
      note=notes[tunevoice[2]+octave];
      break;
    case 10:
      note=notes[tunevoice[3]+octave];
      break;
  }
  set_parameters(0,note,70000,10,0);
}

inline int get_next_overall() {
int total,i;

  beatc++;
  if(beatc>tunespeed) {
    beatc=0;
    next_beat();
  }
  total=0;
  for(i=0;i<VOICES;i++) total+=get_next_byte(i);
  return total/VOICES;
}

void initialise_types() {
int i,j,oct;

  for(i=0;i<512;i++) {
    triangle[i]=i-128;
    if(i>255) triangle[i]=383-i;
    sawtooth[i]=i/2-128;
    sine[i]=(char)(127.0*sin((((float)i)/256.0)*3.141593));
  }
  notes[0]=3.037467; /* C1 */
  notes[1]=3.218084;
  notes[2]=3.409441;
  notes[3]=3.612177;
  notes[4]=3.826968;
  notes[5]=4.054532;
  notes[6]=4.295627;
  notes[7]=4.551058;
  notes[8]=4.821678;
  notes[9]=5.10839;  /* A1 */
  notes[10]=5.412151;
  notes[11]=5.733974;
  oct=2;
  for(i=1;i<6;i++) {
    for(j=0;j<12;j++) {
      notes[12*i+j]=notes[j]*oct;
    }
    oct=oct*2;
  }
  /* note*2^(octave-1) */
  /* Highest note is E7 */
  currentune=-1;
}

void fill_music_buffer(int musicfd) {
char buf[235];
int i;

  for(i=0;i<235;i++) buf[i]=get_next_overall();
  write(musicfd,buf,235);
}
