#ifndef __SPS2WRAP_H__
#define __SPS2WRAP_H__

#include <sps2lib.h>
#include <sps2util.h>
#include <sps2tags.h>
#include "singleton.h"

#define SPS2Manager CSPS2Manager::GetSingleton()

class CDMAMem;

class CSPS2Manager : public CSingleton<CSPS2Manager>
{
	public:
		CSPS2Manager();
		~CSPS2Manager();

		void Initialise(const int iAllocPages);

		void BeginScene(int R, int G, int B);
		void EndScene();

		void Allocate(CDMAMem & Memory, int iQuadWords);

		void FlushCache();

		int GetScreenWidth()
		{ return sps2UScreenGetWidth(); }

		int GetScreenHeight()
		{ return sps2UScreenGetHeight(); }
		
	private:
		int m_iSPS2Desc;

		sps2Memory_t * m_pMemCached, 
			     * m_pMemUncached;
		
		int * m_iaFreeMemInPage;
		int m_iNumPages;
};

inline void CSPS2Manager::EndScene()
{
	sps2UScreenSwap();
}

inline void CSPS2Manager::FlushCache()
{
	sps2FlushCache(m_iSPS2Desc);
}

class CDMAMem
{
	public:
		CDMAMem()
		:m_iAmount(0), m_pCached(NULL), 
		 m_pUncached(NULL), m_iPhysAddr(0)
		 {};
		~CDMAMem(){};

		CDMAMem(int iQWC)
		{
			SPS2Manager.Allocate(*this, iQWC);
		}

		void Set(int Amount, void * pCached, void * pUncached, int PhysAddr)
		{
			m_iAmount = Amount;
			m_pCached = pCached;
			m_pUncached = pUncached;
			m_iPhysAddr = PhysAddr;
		}

		void * GetCached()
		{ return m_pCached; }

		void * GetUncached()
		{ return m_pUncached; }

		int GetPhysicalAddr()
		{ return m_iPhysAddr; }
		
		int GetSize()
		{ return m_iAmount; }
		
	protected:
		int m_iAmount;
		void * m_pCached,
		     * m_pUncached;
		int m_iPhysAddr;
};

#endif
